#
#  Copyright © 2021 Ingram Micro Inc. All rights reserved.
#

from dj_rql.drf._utils import get_query
from dj_rql.drf.backend import FilterCache, RQLFilterBackend
from dj_rql.drf.paginations import RQLContentRangeLimitOffsetPagination, RQLLimitOffsetPagination


__all__ = [
    'get_query',
    'FilterCache',
    'RQLContentRangeLimitOffsetPagination',
    'RQLFilterBackend',
    'RQLLimitOffsetPagination',
]



import os
import sys
from datetime import datetime

import django

from setuptools_scm import get_version


sys.path.insert(0, os.path.abspath('..'))
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.dj_rf.settings'
django.setup()


# -- Project information -----------------------------------------------------

project = 'django-rql'
copyright = '{0}, CloudBlue Inc.'.format(datetime.now().year)
author = 'CloudBlue'

# The full version, including alpha/beta/rc tags
release = get_version(root='..', relative_to=__file__)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx_copybutton',
]

autosectionlabel_prefix_document = True
autodoc_member_order = 'bysource'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



#
#  Copyright © 2021 Ingram Micro Inc. All rights reserved.
#

from copy import copy
from numbers import Number

from dj_rql.constants import (
    FilterLookups, FilterTypes, RQL_NULL, RQL_ORDERING_OPERATOR, RQL_SEARCH_PARAM,
)


class RQLFilterDescriptionTemplate:
    BASE_TEMPLATE = '**{description}**\n\nlookups: {lookups}'
    DEFAULT_DESCRIPTION = 'Filter for: {description}'
    IN_PLACE_RENDERERS = (
        '_render_search_inplace',
        '_render_ordering_inplace',
        '_render_null_inplace',
        '_render_default_inplace',
    )

    @classmethod
    def render(cls, filter_item, filter_instance):
        """
        :param dict filter_item: Extended Filter item
        :param dj_rql.filter_cls.RQLFilterClass filter_instance: Instance of Filter Class
        :return: Rendered description for filter item
        :rtype: str
        """
        result = cls._render_base(filter_item, filter_instance)

        for render_func in cls.IN_PLACE_RENDERERS:
            result = getattr(cls, render_func)(result, filter_item, filter_instance)

        return result

    @classmethod
    def _render_base(cls, filter_item, filter_instance):
        sorted_lookups = (
            FilterLookups.EQ,
            FilterLookups.NE,
            FilterLookups.GE,
            FilterLookups.GT,
            FilterLookups.LE,
            FilterLookups.LT,
            FilterLookups.LIKE,
            FilterLookups.I_LIKE,
            FilterLookups.NULL,
            FilterLookups.IN,
            FilterLookups.OUT,
        )
        filter_lookups = filter_item.get('lookups', {})

        description = filter_item['oa']['description']
        if not description:
            description = cls.DEFAULT_DESCRIPTION.format(description=filter_item['oa']['name'])

        return cls.BASE_TEMPLATE.format(
            description=description,
            lookups=', '.join(fl for fl in sorted_lookups if fl in filter_lookups),
        )

    @classmethod
    def _render_search_inplace(cls, base, filter_item, filter_instance):
        if filter_item['oa']['name'] not in filter_instance.search_filters:
            return base

        return cls._render_common_key_inplace(base, RQL_SEARCH_PARAM, 'true')

    @classmethod
    def _render_ordering_inplace(cls, base, filter_item, filter_instance):
        if filter_item['oa']['name'] not in filter_instance.ordering_filters:
            return base

        return cls._render_common_key_inplace(base, RQL_ORDERING_OPERATOR, 'true')

    @classmethod
    def _render_null_inplace(cls, base, filter_item, filter_instance):
        null_values = filter_item.get('null_values', {RQL_NULL})

        if FilterLookups.NULL in filter_item.get('lookups', {}) and null_values != {RQL_NULL}:
            return cls._render_common_key_inplace(
                base, 'null', ', '.join(sorted(null_values)),
            )

        return base

    @classmethod
    def _render_default_inplace(cls, base, filter_item, filter_instance):
        if filter_item.get('hidden', False):
            return cls._render_common_key_inplace(base, 'default', '**hidden**')

        return base

    @classmethod
    def _render_common_key_inplace(cls, base, key, value):
        return '{base}\n{key}: {value}'.format(base=base, key=key, value=value)


class RQLFilterClassSpecification:
    FIELD_DESCRIPTION_TEMPLATE = RQLFilterDescriptionTemplate

    @classmethod
    def get(cls, filter_instance):
        """
        Returns OpenAPI specification for filters.
        Filter sorting is alphabetic with deprecated filters in the end.

        :param dj_rql.filter_cls.RQLFilterClass filter_instance: Instance of Filter Class
        :return: OpenAPI compatible specification of Filter Class Filters
        :rtype: list of dict
        """
        extended_filter_items = {}
        common_filter_names, deprecated_filter_names = [], []

        for filter_name, filter_item in filter_instance.filters.items():
            f_item = filter_item[0] if isinstance(filter_item, list) else filter_item
            openapi_data = cls._get_filter_item_openapi_data(filter_name, f_item)

            if not openapi_data['hidden']:
                extended_filter_items[filter_name] = copy(f_item)
                extended_filter_items[filter_name]['oa'] = openapi_data

                if not openapi_data['deprecated']:
                    common_filter_names.append(filter_name)
                else:
                    deprecated_filter_names.append(filter_name)

        result = []
        for filter_name in sorted(common_filter_names) + sorted(deprecated_filter_names):
            filter_item = extended_filter_items[filter_name]

            filter_spec = cls.get_for_field(filter_item, filter_instance)
            if not filter_spec:
                filter_spec = cls._get_default_for_field(filter_item, filter_instance)

            result.append(filter_spec)

        return result

    @classmethod
    def get_for_field(cls, filter_item, filter_instance):
        """ This method can be overridden to support custom specs for certain filters.

        :param dict filter_item: Extended Filter Item
        :param dj_rql.filter_cls.RQLFilterClass filter_instance: Instance of Filter Class
        :rtype: dict or None
        """
        pass

    @classmethod
    def _get_default_for_field(cls, filter_item, filter_instance):
        return {
            'name': filter_item['oa']['name'],
            'description': cls.FIELD_DESCRIPTION_TEMPLATE.render(filter_item, filter_instance),
            'in': 'query',
            'required': filter_item['oa']['required'],
            'deprecated': filter_item['oa']['deprecated'],
            'schema': cls._get_schema_for_field(filter_item, filter_instance),
        }

    @classmethod
    def _get_schema_for_field(cls, filter_item, filter_instance):
        result = {}

        if filter_item['oa']['type']:
            result = {'type': filter_item['oa']['type']}

            if filter_item['oa']['format']:
                result['format'] = filter_item['oa']['format']

        if not result and ('field' not in filter_item):
            return {'type': 'string'}

        field = filter_item['field']
        if not result:
            field_type_oa_type_mapper = {
                FilterTypes.STRING: {'type': 'string'},
                FilterTypes.INT: {'type': 'integer'},
                FilterTypes.DECIMAL: {'type': 'number', 'format': 'double'},
                FilterTypes.FLOAT: {'type': 'number', 'format': 'float'},
                FilterTypes.DATETIME: {'type': 'string', 'format': 'date-time'},
                FilterTypes.DATE: {'type': 'string', 'format': 'date'},
                FilterTypes.BOOLEAN: {'type': 'boolean'},
            }
            result = field_type_oa_type_mapper[FilterTypes.field_filter_type(field)]

        choices = getattr(field, 'choices', None)
        if choices:
            if type(choices).__name__ == 'Choices' or isinstance(choices[0], tuple):
                use_repr = filter_item.get('use_repr', False)
                enum = [choice[int(use_repr)] for choice in choices]
            else:
                enum = list(choices)
            result['enum'] = enum

            if isinstance(enum[0], Number):
                result['type'] = 'integer'
            else:
                result['type'] = 'string'

        return result

    @classmethod
    def _get_filter_item_openapi_data(cls, filter_name, filter_item):
        openapi_data = copy(filter_item.get('openapi', {}))

        openapi_data['description'] = openapi_data.get('description')
        openapi_data['hidden'] = openapi_data.get('hidden', False)
        openapi_data['required'] = openapi_data.get('required', False)
        openapi_data['deprecated'] = openapi_data.get('deprecated', False)
        openapi_data['name'] = filter_name
        openapi_data['type'] = openapi_data.get('type')
        openapi_data['format'] = openapi_data.get('format')

        return openapi_data



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#

# flake8: noqa
"""
RQL EBNF LALR(1) compatible grammar for standard Lark lexer.

Notes:
    - rules are lowercase. They only consist of values, other rules and terminals.
    - TERMINALS are uppercase. They only consist of values and other terminals.
"""

RQL_GRAMMAR = r"""
start: term?

term: expr_term
    | logical
    
expr_term: comp
    | listing
    | searching
    | ordering
    | select
    | _L_BRACE expr_term _R_BRACE
    
logical: and_op
    | or_op
    | not_op
    
and_op: _and 
    | _L_BRACE _and _R_BRACE

_and: _AND _logical_exp
    | term "&" term
    | term _COMMA term

or_op: _or
    | _L_BRACE _or _R_BRACE

_or: _OR _logical_exp
    | _L_BRACE term "|" term _R_BRACE
    | _L_BRACE term ";" term _R_BRACE

_logical_exp: _L_BRACE expr_term (_COMMA expr_term)+ _R_BRACE

not_op: _NOT _L_BRACE expr_term _R_BRACE

comp: comp_term _L_BRACE prop _COMMA val _R_BRACE
    | prop _EQUALITY comp_term _EQUALITY val
    | prop _EQUALITY val
    
listing: list_term _L_BRACE prop _COMMA _L_BRACE val (_COMMA val)* _R_BRACE _R_BRACE
searching: search_term _L_BRACE prop _COMMA val _R_BRACE

ordering: ordering_term _signed_props
select: select_term _signed_props
_signed_props: _L_BRACE _R_BRACE
    | _L_BRACE sign_prop (_COMMA sign_prop)* _R_BRACE
    
val: prop
    | QUOTED_VAL
    | UNQUOTED_VAL
    
prop: comp_term
    | logical_term
    | list_term
    | search_term
    | ordering_term
    | select_term
    | PROP
    
!sign_prop: ["+"|"-"] prop
    
!comp_term: "eq" | "ne" | "gt" | "ge" | "lt" | "le"
!logical_term: _AND | _OR | _NOT
!list_term: "in" | "out"
!search_term: "like" | "ilike"
!ordering_term: "ordering"
!select_term: "select"

    
PROP: /[a-zA-Z]/ /[\w\-\.]/*
QUOTED_VAL: /"[^"]*"/
    | /'[^']*'/
UNQUOTED_VAL: NULL
    | EMPTY
    | /[\w\-\*\+\\]/ /[\w\.\s\-\:\+\@\*\\]/*

EMPTY: "empty()"
NULL: "null()"

_AND: "and"
_OR: "or"
_NOT: "not"

_COMMA: ","
_L_BRACE: "("
_R_BRACE: ")"
_EQUALITY: "="
"""



#
#  Copyright © 2021 Ingram Micro Inc. All rights reserved.
#

from collections import namedtuple

from django.db.models import Prefetch


class DBOptimization:
    def __init__(self, *relations, **kwargs):
        assert relations, 'At least one optimization must be specified.'

        self._relations = relations
        self._extensions = kwargs

    def rebuild(self, parent_optimization=None):
        return self.__class__(*self._relations, **self._extensions)

    @property
    def main_relation(self):
        return self._relations[0]

    @property
    def extensions(self):
        return self._extensions

    @property
    def relations(self):
        return self._relations

    def apply(self, queryset):
        raise NotImplementedError


class Annotation(DBOptimization):
    def __init__(self, **kwargs):
        super(Annotation, self).__init__('None', **kwargs)

    def rebuild(self, parent_optimization=None):
        return self

    def apply(self, queryset):
        return queryset.annotate(**self._extensions)


class SelectRelated(DBOptimization):
    """Apply a ``select_related`` optimization to the queryset."""
    def apply(self, queryset):
        return queryset.select_related(*self._relations)


class PrefetchRelated(DBOptimization):
    """Apply a ``prefetch_related`` optimization to the queryset."""
    def apply(self, queryset):
        return queryset.prefetch_related(*self._relations)


_ParentData = namedtuple('_ParentData', 'parent relation type')


class _NestedOptimizationMixin:
    _SR = 0
    _PR = 1

    def rebuild(self, parent_optimization=None):
        if not parent_optimization or isinstance(parent_optimization, Annotation):
            return super(_NestedOptimizationMixin, self).rebuild()

        real_parent_optimization = parent_optimization
        while isinstance(real_parent_optimization, Chain):
            real_parent_optimization = real_parent_optimization.main_relation

        parent_relation = real_parent_optimization.main_relation
        assert isinstance(parent_relation, str), 'Only simple parent relations are supported.'

        if isinstance(real_parent_optimization, PrefetchRelated):
            parent_type = self._PR
        else:
            parent_type = self._SR

        return self._rebuild_nested(
            _ParentData(real_parent_optimization, parent_relation, parent_type),
        )

    def _rebuild_nested(self, parent_data):
        """
        :param _ParentData parent_data:
        """
        raise NotImplementedError

    @staticmethod
    def _join_relation(parent_relation, relation):
        return '{0}__{1}'.format(parent_relation, relation)


class NestedPrefetchRelated(_NestedOptimizationMixin, PrefetchRelated):
    def _rebuild_nested(self, parent_data):
        rebuilt_relations = []

        for relation in self._relations:
            if isinstance(relation, Prefetch):
                rebuilt_relations.append(
                    Prefetch(
                        self._join_relation(parent_data.relation, relation.prefetch_to),
                        queryset=relation.queryset,
                        to_attr=relation.to_attr,
                    ),
                )
            else:
                rebuilt_relations.append(self._join_relation(parent_data.relation, relation))

        return self.__class__(*rebuilt_relations, **self._extensions)


class NestedSelectRelated(_NestedOptimizationMixin, SelectRelated):
    def _rebuild_nested(self, parent_data):
        optimization_cls = NSR if parent_data.type == self._SR else NPR
        rebuilt_relations = [self._join_relation(parent_data.relation, r) for r in self._relations]

        return optimization_cls(*rebuilt_relations, **self._extensions)


class Chain(_NestedOptimizationMixin, DBOptimization):
    def __init__(self, *relations, **extensions):
        e = 'Wrong Chain() optimization configuration.'
        assert all(isinstance(rel, DBOptimization) for rel in relations), e

        super(Chain, self).__init__(*relations, **extensions)

    def _rebuild_nested(self, parent_data):
        rebuilt_relations = [r.rebuild(parent_data.parent) for r in self.relations]

        return self.__class__(*rebuilt_relations, **self.extensions)

    def apply(self, queryset):
        for opt in self._relations:
            queryset = opt.apply(queryset)

        return queryset


AN = Annotation
SR = SelectRelated
NSR = NestedSelectRelated
PR = PrefetchRelated
NPR = NestedPrefetchRelated
CH = Chain



#
#  Copyright © 2021 Ingram Micro Inc. All rights reserved.
#

from dj_rql.exceptions import RQLFilterParsingError
from dj_rql.grammar import RQL_GRAMMAR

from lark import Lark
from lark.exceptions import LarkError


class RQLLarkParser(Lark):
    def parse_query(self, query):
        try:
            rql_ast = RQLParser.parse(query)
            return rql_ast
        except LarkError:
            raise RQLFilterParsingError()


RQLParser = RQLLarkParser(RQL_GRAMMAR, parser='lalr', start='start')



#
#  Copyright © 2021 Ingram Micro Inc. All rights reserved.
#

from collections import Counter

from dj_rql.constants import (
    ComparisonOperators as CO,
    DjangoLookups as DJL,
    FilterTypes,
    RQL_ANY_SYMBOL,
    RQL_FALSE,
    RQL_LIMIT_PARAM,
    RQL_NULL,
    RQL_OFFSET_PARAM,
    RQL_ORDERING_OPERATOR,
    RQL_TRUE,
    SearchOperators as SO,
)
from dj_rql.drf._utils import get_query
from dj_rql.drf.backend import RQLFilterBackend
from dj_rql.exceptions import RQLFilterParsingError


class CompatibilityRQLFilterBackend(RQLFilterBackend):
    """
    If there is necessity to apply RQL filters to a production API, which was working on other
    filter backend (without raising API version number or losing compatibility), this base
    compatibility DRF backend must be inherited from.
    """
    @classmethod
    def get_query(cls, filter_instance, request, view):
        try:
            query_string = cls.modify_initial_query(filter_instance, request, get_query(request))

            if not cls.is_old_syntax(filter_instance, request, query_string):
                return query_string

            else:
                return cls.get_rql_query(filter_instance, request, query_string)
        except Exception:
            raise RQLFilterParsingError()

    @classmethod
    def modify_initial_query(cls, filter_instance, request, query_string):
        return query_string

    @classmethod
    def is_old_syntax(cls, filter_instance, request, query_string):
        raise NotImplementedError

    @classmethod
    def get_rql_query(cls, filter_instance, request, query_string):
        raise NotImplementedError


class DjangoFiltersRQLFilterBackend(CompatibilityRQLFilterBackend):
    """
    DRF Backend, that automatically converts Django Filter specific queries to correct RQL queries.

    IMPORTANT NOTES:
        * `;` separation is context-based;
        * MultipleChoiceFilter works by OR logic in RQL.

    Currently NOT SUPPORTED:
        * range fields and filters;
        * regex and iregex conversion;
        * OrderingFilter;

        * ?&& syntax in queries;
        * etc.
    """
    RESERVED_ORDERING_WORDS = {'order_by', 'ordering'}

    _POSSIBLE_DF_LOOKUPS = DJL.all()
    _RQL_COMPARISON_OPERATORS = {CO.EQ, CO.NE, CO.LE, CO.GE, CO.LT, CO.GT}
    _IMPOSSIBLE_PROP_SYMBOLS = {'(', ',', ')', ' ', "'", '"'}

    @classmethod
    def is_old_syntax(cls, filter_instance, request, query_string):
        if not query_string.strip():
            return False

        if query_string[-1] == '&':
            return True

        qp_all_filters = set()
        qp_old_filters = set()
        query_params = request.query_params
        for filter_name in query_params.keys():
            result = cls._filter_has_old_syntax(filter_name, query_params)
            if result is not None:
                return result

            qp_all_filters.add(filter_name)
            if cls._is_old_style_filter(filter_name):
                lookup = cls._get_filter_and_lookup(filter_name)[-1]
                if lookup in (DJL.REGEX, DJL.I_REGEX):
                    cls._conversion_error()

                qp_old_filters.add(filter_name)

        if not qp_all_filters.isdisjoint(cls.RESERVED_ORDERING_WORDS):
            return True

        if not qp_old_filters:
            return False

        if not qp_old_filters - cls._get_filters_similar_to_old_syntax(filter_instance):
            return False

        return True

    @classmethod
    def _filter_has_old_syntax(cls, filter_name, query_params):
        has_select = not cls._is_select_in_filter(filter_name)
        if has_select and (not set(Counter(filter_name)).isdisjoint(cls._IMPOSSIBLE_PROP_SYMBOLS)):
            return False

        return cls._filter_value_has_old_syntax(filter_name, query_params, has_select)

    @classmethod
    def _filter_value_has_old_syntax(cls, filter_name, query_params, has_select):
        for v in query_params.getlist(filter_name):
            if has_select and not v:
                return True

            if v in ('True', 'False'):
                return True

            is_old = cls._filter_value_has_old_syntax_by_special_chars(v)
            if is_old is not None:
                return is_old

    @classmethod
    def _filter_value_has_old_syntax_by_special_chars(cls, value):
        vc = Counter(value)
        no_quotes = not (vc.get('"', 0) > 1 or vc.get("'", 0) > 1)
        if vc.get(' ') and no_quotes:
            return True

        number_of_eqs = vc.get('=', 0)
        if number_of_eqs >= 1 and vc.get('(', 0) == 0 and vc.get(';'):
            return True

        if number_of_eqs and no_quotes:
            return False

        if len(value) > 2 and value[2] == '=' and value[:2] in cls._RQL_COMPARISON_OPERATORS:
            return False

    @classmethod
    def get_rql_query(cls, filter_instance, request, query_string):
        filter_value_pairs = []

        for filter_name in request.query_params.keys():
            if cls._is_select_in_filter(filter_name):
                filter_value_pairs.append(filter_name)
                continue

            one_filter_value_pairs = []
            for value in request.query_params.getlist(filter_name):
                name_value_pair = cls._get_one_filter_value_pair(
                    filter_instance, filter_name, value,
                )
                if name_value_pair is not None:
                    one_filter_value_pairs.append(name_value_pair)

            if one_filter_value_pairs:
                filter_value_pairs.append('&'.join(one_filter_value_pairs))

        return '&'.join(filter_value_pairs) if filter_value_pairs else ''

    @classmethod
    def _get_one_filter_value_pair(cls, filter_instance, filter_name, value):
        if not value:
            return

        if filter_name in (RQL_LIMIT_PARAM, RQL_OFFSET_PARAM):
            return '{0}={1}'.format(filter_name, value)

        if filter_name in cls.RESERVED_ORDERING_WORDS:
            return '{0}({1})'.format(RQL_ORDERING_OPERATOR, value)

        f_item = filter_instance.get_filter_base_item(filter_name)
        is_nc_item = f_item and (not f_item.get('custom', False))
        if is_nc_item and FilterTypes.field_filter_type(f_item['field']) == FilterTypes.BOOLEAN:
            value = cls._convert_bool_value(value)

        if not cls._is_old_style_filter(filter_name):
            return '{0}={1}'.format(filter_name, cls._add_quotes_to_value(value))

        return cls._convert_filter_to_rql(filter_name, value)

    @staticmethod
    def _is_select_in_filter(filter_name):
        return 'select(' in filter_name

    @classmethod
    def _convert_filter_to_rql(cls, filter_name, value):
        filter_base, lookup = cls._get_filter_and_lookup(filter_name)

        if lookup == DJL.IN:
            return 'in({0},({1}))'.format(
                filter_base, ','.join(cls._add_quotes_to_value(v) for v in value.split(',') if v),
            )

        if lookup == DJL.NULL:
            operator = CO.EQ if cls._convert_bool_value(value) == 'true' else CO.NE
            return '{0}={1}={2}'.format(filter_base, operator, RQL_NULL)

        if lookup in (DJL.GT, DJL.GTE, DJL.LT, DJL.LTE):
            if lookup == DJL.GTE:
                operator = CO.GE
            elif lookup == DJL.LTE:
                operator = CO.LE
            else:
                operator = lookup
            return '{0}={1}={2}'.format(filter_base, operator, value)

        operator = SO.I_LIKE if lookup[0] == 'i' else SO.LIKE

        lookups = (DJL.CONTAINS, DJL.I_CONTAINS, DJL.ENDSWITH, DJL.I_ENDSWITH)
        if lookup in lookups and value[0] != RQL_ANY_SYMBOL:
            value = RQL_ANY_SYMBOL + value

        lookups = (DJL.CONTAINS, DJL.I_CONTAINS, DJL.STARTSWITH, DJL.I_STARTSWITH)
        if lookup in lookups and value[-1] != RQL_ANY_SYMBOL:
            value += RQL_ANY_SYMBOL

        return '{0}({1},{2})'.format(operator, filter_base, cls._add_quotes_to_value(value))

    @classmethod
    def _convert_bool_value(cls, value):
        if value in ('True', 'true', '1'):
            return RQL_TRUE
        elif value in ('False', 'false', '0'):
            return RQL_FALSE

        cls._conversion_error()

    @classmethod
    def _add_quotes_to_value(cls, value):
        for quote in ('"', "'"):
            if quote not in value:
                return '{q}{0}{q}'.format(value, q=quote)

        cls._conversion_error()

    @staticmethod
    def _conversion_error():
        raise RQLFilterParsingError()

    @classmethod
    def _get_filters_similar_to_old_syntax(cls, filter_instance):
        old_syntax_filters = getattr(filter_instance, 'old_syntax_filters', None)
        if old_syntax_filters:
            return old_syntax_filters

        similar_to_old_syntax_filters = set()
        for filter_name in filter_instance.filters.keys():
            if cls._is_old_style_filter(filter_name):
                similar_to_old_syntax_filters.add(filter_name)

        filter_instance.old_syntax_filters = similar_to_old_syntax_filters
        return similar_to_old_syntax_filters

    @classmethod
    def _is_old_style_filter(cls, filter_name):
        return cls._get_filter_and_lookup(filter_name)[-1] in cls._POSSIBLE_DF_LOOKUPS

    @classmethod
    def _get_filter_and_lookup(cls, filter_name):
        return filter_name.rsplit('__', 1)



#
#  Copyright © 2021 Ingram Micro Inc. All rights reserved.
#

from dj_rql.drf._utils import get_query
from dj_rql.exceptions import RQLFilterParsingError
from dj_rql.parser import RQLParser
from dj_rql.transformer import RQLLimitOffsetTransformer

from lark.exceptions import LarkError

from rest_framework.pagination import LimitOffsetPagination, _positive_int
from rest_framework.response import Response


class RQLLimitOffsetPagination(LimitOffsetPagination):
    """ RQL limit offset pagination. """
    def __init__(self, *args, **kwargs):
        super(RQLLimitOffsetPagination, self).__init__(*args, **kwargs)

        self._rql_limit = None
        self._rql_offset = None

    def get_paginated_response_schema(self, schema):
        return schema

    def paginate_queryset(self, queryset, request, view=None):
        rql_ast = None
        try:
            rql_ast = request.rql_ast
        except AttributeError:
            query = get_query(request)
            if query:
                rql_ast = RQLParser.parse_query(query)

        if rql_ast is not None:
            try:
                self._rql_limit, self._rql_offset = RQLLimitOffsetTransformer().transform(rql_ast)
            except LarkError:
                raise RQLFilterParsingError(details={
                    'error': 'Limit and offset are set incorrectly.',
                })

        if self.get_limit() == 0:
            self.count = self.get_count(queryset)
            self.offset = 0
            return []

        return super(RQLLimitOffsetPagination, self).paginate_queryset(queryset, request, view)

    def get_limit(self, *args):
        if self._rql_limit is not None:
            try:
                return _positive_int(self._rql_limit, strict=False, cutoff=self.max_limit)
            except ValueError:
                pass
        return self.default_limit

    def get_offset(self, *args):
        if self._rql_offset is not None:
            try:
                return _positive_int(self._rql_offset)
            except ValueError:
                pass
        return 0


class RQLContentRangeLimitOffsetPagination(RQLLimitOffsetPagination):
    """ RQL RFC2616 limit offset pagination.

    Examples:
        Response

        200 OK
        Content-Range: items <FIRST>-<LAST>/<TOTAL>
    """

    def get_paginated_response(self, data):
        length = len(data) - 1 if data else 0
        content_range = 'items {0}-{1}/{2}'.format(
            self.offset, self.offset + length, self.count,
        )
        return Response(data, headers={'Content-Range': content_range})



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#


class OptimizationArgs:
    def __init__(self, queryset, select_data, filter_tree, filter_node=None, filter_path=None):
        """
        :param django.db.models.QuerySet queryset: QuerySet for optimization
        :param Dict[str, bool] select_data: Storage of selected/deselected fields (filters)
        :param Dict[str, dict] filter_tree: Detailed tree structure of filter items
        :param dict or None filter_node: Field (filter node) on which optimization is applied
        :param str or None filter_path: Full RQL field path (full filter name)
        """
        self.queryset = queryset
        self.select_data = select_data
        self.filter_tree = filter_tree

        self.filter_node = filter_node
        self.filter_path = filter_path


class FilterArgs:
    def __init__(self, filter_name, operator, str_value, list_operator=None, **kwargs):
        """
        :param str filter_name: Full filter name (f.e. ns1.ns2.filter1)
        :param str operator: RQL operator (f.e. eq, like, etc.)
        :param str str_value: Raw value from RQL query
        :param str or None list_operator: This is filled only if operation is done within IN or OUT
        :param dict kwargs: Other auxiliary data (f.e. to ease custom filtering)
        """
        self.filter_name = filter_name
        self.operator = operator
        self.str_value = str_value
        self.list_operator = list_operator

        self.filter_lookup = kwargs.get('filter_lookup')
        self.django_lookup = kwargs.get('django_lookup')



#
#  Copyright © 2020 Ingram Micro Inc. All rights reserved.
#


class RQLFilterError(Exception):
    """ Base class for RQL errors. """
    MESSAGE = 'RQL Filtering error.'

    def __init__(self, details=None):
        super(RQLFilterError, self).__init__(self.MESSAGE)
        self.details = details


class RQLFilterParsingError(RQLFilterError):
    """ Parsing errors are raised only at query parsing time. """
    MESSAGE = 'RQL Parsing error.'

    def __init__(self, **kwargs):
        kwargs.setdefault('details', {'error': 'Bad filter query.'})
        super().__init__(**kwargs)


class RQLFilterLookupError(RQLFilterError):
    """ Lookup error is raised when provided lookup is not supported by the associated filter. """
    MESSAGE = 'RQL Lookup error.'


class RQLFilterValueError(RQLFilterError):
    """ Value error is raised when provided values can't be converted to DB field types. """
    MESSAGE = 'RQL Value error.'
